@@ -1,12 +1,13 @@
 /*! \file
  * Access functions for  ~/.dillo/dpi_socket_dir.
- * The most useful function for dillo is a_Dpi_srs, it returns 
+ * The most useful function for dillo is a_Dpi_srs, it returns
  * the full path to the dpid service request socket.
  */
 
 #ifndef DPI_H
 #define DPI_H
 
+#include <config.h>
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -44,13 +45,6 @@ enum {
    REGISTER_SERVICE_CMD /* "register_service" */
 };
 
-/*!
- * Dpi error message tags
- */
-#define tag_prserr "<dpi cmd='DpiError' msg='Failed to parse request'>"
-#define tag_cmderr "<dpi cmd='DpiError' msg='Unknown command'>"
-#define tag_dpi_unavailerr\
- "<dpi cmd='DpiError' msg='Plugin currently unavailable'>"
 
 char *a_Dpi_sockdir_file(void);
 
