@@ -23,7 +23,6 @@
 #include <sys/stat.h>
 #include <sys/uio.h>
 #include <sys/socket.h>
-#include <glib.h>
 #include <gdk/gdk.h>
 #include "../msg.h"
 #include "../chain.h"
@@ -50,6 +49,12 @@
 static Klist_t *ValidIOs = NULL; /* Active IOs list. It holds pointers to
                                   * IOData_t structures. */
 
+/*
+ *  Forward declarations
+ */
+void a_IO_ccc(int Op, int Branch, int Dir, ChainLink *Info,
+              void *Data1, void *Data2);
+
 /*
  * Debug procedure...
 
@@ -120,23 +125,23 @@ void a_IO_add_buf(IOData_t *io, void *Buf, size_t BufSize)
       return;
    }
 
-   offset = (io->Buf) ? io->Buf - io->BufStart : 0;
+   offset = (io->Buf) ? (gchar*)io->Buf - (gchar*)io->BufStart : 0;
    io->BufStart = g_realloc(io->BufStart, offset + io->BufSize + BufSize);
-   memcpy(io->BufStart + offset + io->BufSize, Buf, BufSize);
-   io->Buf = io->BufStart + offset;
+   memcpy((gchar*)io->BufStart + offset + io->BufSize, Buf, BufSize);
+   io->Buf = (gchar*)io->BufStart + offset;
    io->BufSize += BufSize;
    io->Flags |= IOFlag_FreeIOBuf;
 }
 
 /*
  * Return transfer buffer.
- */
+ *
 void a_IO_get_buf(IOData_t *io, void **Buf, size_t *BufSize)
 {
    *Buf = io->Buf;
    *BufSize = io->BufSize;
 }
-
+ */
 
 typedef struct {
    pthread_t thrID;        /* Thread id */
@@ -144,6 +149,8 @@ typedef struct {
    pthread_cond_t cond;
 
    gint FD;
+   void *Key;
+
    void *Buf1Start;
    void *Buf1;
    size_t Buf1Size;
@@ -151,6 +158,7 @@ typedef struct {
    size_t Buf2Size;
 
    gint LastChunk;
+   gint CloseFD;
    gint Done;
 } thr_data_t;
 
@@ -162,7 +170,7 @@ static gint ThrDataMax = 8;
 /*
  * Create a new ThrData node
  */
-static thr_data_t *IO_thrdata_new(gint FD)
+static thr_data_t *IO_thrdata_new(gint FD, void *Key)
 {
    thr_data_t *data;
 
@@ -170,9 +178,11 @@ static thr_data_t *IO_thrdata_new(gint FD)
    pthread_mutex_init(&data->mut, NULL);
    pthread_cond_init(&data->cond, NULL);
    data->FD = FD;
+   data->Key = Key;
    data->Buf1Start = data->Buf1 = data->Buf2 = NULL;
    data->Buf1Size = data->Buf2Size = 0;
    data->LastChunk = 0;
+   data->CloseFD = 0;
    data->Done = 0;
    return data;
 }
@@ -182,28 +192,41 @@ static thr_data_t *IO_thrdata_new(gint FD)
  */
 static void IO_thrdata_free(thr_data_t *td)
 {
+   /* EBUSY should not happen: IO_thrdata_free is called after the
+    * thread is done */
+   while (pthread_cond_destroy(&td->cond) == EBUSY) {
+      g_warning("IO_thrdata_free: EBUSY\n");
+      if (pthread_cancel(td->thrID) == ESRCH)
+         break;
+   }
+
    pthread_mutex_destroy(&td->mut);
-   pthread_cond_destroy(&td->cond);
    g_free(td);
 }
 
 /*
  * Search data node for a FD
- * (This also frees already closed data)
+ *  - The FD is searched using a Key because using the FD itself may fail
+ *  due to a race condition between the FD close and its reuse.
+ *  - This function also frees already closed data.
  */
-static thr_data_t *IO_thrdata_find(gint FD)
+static thr_data_t *IO_thrdata_find(void *Key)
 {
    gint i, idx = -1;
 
+   _MSG("TL(%p): [", Key);
    for (i = 0; i < ThrDataSize; ++i) {
+      _MSG(" %d%s", ThrData[i]->FD, ThrData[i]->Done ? "D" : "");
       if (ThrData[i]->Done) {
          IO_thrdata_free(ThrData[i]);
          a_List_remove(ThrData, i, ThrDataSize);
          --i;
-      } else if (ThrData[i]->FD == FD) {
+      } else if (ThrData[i]->Key == Key) {
          idx = i;
+         _MSG("*");
       }
    }
+   _MSG("]\n");
 
    if (idx != -1)
       return ThrData[idx];
@@ -222,9 +245,6 @@ static void *IO_write_chunk(void *ptr)
    ssize_t St;
    gint st, done;
 
-   /* switch to detached state */
-   pthread_detach(data->thrID);
-
    _MSG("thr::\n");
    _MSG("      [thrID:%lu]\n", (gulong)data->thrID);
 
@@ -268,7 +288,6 @@ static void *IO_write_chunk(void *ptr)
                data->Buf1Size -= St;
             } else {
                /* All data in buffer written */
-               //sleep(1);
                done = 1;
             }
          }
@@ -282,9 +301,13 @@ static void *IO_write_chunk(void *ptr)
       if (data->Buf2)
          continue;
       else if (data->LastChunk) {
-         do
-            st = close(data->FD);
-         while (st < 0 && errno == EINTR);
+         /* Only pipes are closed, sockets are left for the server to close */
+         if (data->CloseFD) {
+            do
+               st = close(data->FD);
+            while (st < 0 && errno == EINTR);
+         }
+         g_free(data->Buf1Start);
          data->Done = 1;
          pthread_mutex_unlock(&data->mut);
          _MSG("thr:: LastChunk:%d Done:%d  --Bailing out!\n",
@@ -301,33 +324,50 @@ static void *IO_write_chunk(void *ptr)
 
 /*
  * Write a data chunk from a pthread
- * (LastChunk -> close tansfer)
+ * (!Buf && !BufSize -> LastChunk -> close tansfer, don't close FD)
+ * (!Buf && BufSize == 1 -> LastChunk -> close tansfer, close FD)
  */
-void a_IO_write_chunk(gint FD, void *Buf, size_t BufSize)
+void a_IO_write_chunk(gint FD, void *Key, void *Buf, size_t BufSize)
 {
    thr_data_t *data;
+   gint new_thread = 0;
+   static pthread_attr_t thrATTR;
+   static gint thrATTRInitialized = 0;
+
+   /* set the thread attribute to the detached state */
+   if (!thrATTRInitialized) {
+      pthread_attr_init(&thrATTR);
+      pthread_attr_setdetachstate(&thrATTR, PTHREAD_CREATE_DETACHED);
+      thrATTRInitialized = 1;
+   }
 
    /* Search data node for this FD */
-   data = IO_thrdata_find(FD);
+   data = IO_thrdata_find(Key);
+   _MSG(" a_IO_write_chunk: data=%p Buf=%p, BufSize=%d, FD=%d\n",
+        data, Buf, BufSize, FD);
 
-   if (!data) {
-      data = IO_thrdata_new(FD);
+   if (!data && Buf) {
+      data = IO_thrdata_new(FD, Key);
       a_List_add(ThrData, ThrDataSize, ThrDataMax);
       ThrData[ThrDataSize] = data;
       ThrDataSize++;
-      pthread_create(&data->thrID, NULL, IO_write_chunk, data);
+      new_thread = 1;
    }
 
    pthread_mutex_lock(&data->mut);
-    if (Buf && BufSize) {
+    if (Buf) {
        data->Buf2 = g_realloc(data->Buf2, data->Buf2Size + BufSize);
-       memcpy(data->Buf2 + data->Buf2Size, Buf, BufSize);
+       memcpy((gchar*)data->Buf2 + data->Buf2Size, Buf, BufSize);
        data->Buf2Size = data->Buf2Size + BufSize;
     } else {
        data->LastChunk = 1;
+       data->CloseFD = (BufSize == 1) ? 1 : 0;
     }
     pthread_cond_signal(&data->cond);
    pthread_mutex_unlock(&data->mut);
+
+   if (new_thread)
+      pthread_create(&data->thrID, &thrATTR, IO_write_chunk, data);
 }
 
 
@@ -362,11 +402,11 @@ static IOData_t *IO_get(gint Key)
  */
 static void IO_free(IOData_t *io)
 {
-  g_return_if_fail(IO_get(io->Key) == NULL);
+   g_return_if_fail(IO_get(io->Key) == NULL);
 
-  if (io->Flags & IOFlag_FreeIOBuf)
-     a_IO_set_buf(io, NULL, 0);
-  g_free(io);
+   if (io->Flags & IOFlag_FreeIOBuf)
+      a_IO_set_buf(io, NULL, 0);
+   g_free(io);
 }
 
 /*
@@ -421,8 +461,8 @@ static gboolean IO_read(IOData_t *io)
 
       St = read(io->FD, io->Buf, io->BufSize);
       io->Status = St;
-      DEBUG_MSG(3, "  IO_read: %s [errno %d] [St %d]\n",
-                g_strerror(errno), errno, St);
+      DEBUG_MSG(3, "  IO_read: %s [errno %d] [St %ld]\n",
+                g_strerror(errno), errno, (glong)St);
 
       if ( St < 0 ) {
          /* Error */
@@ -469,8 +509,8 @@ static gboolean IO_write(IOData_t *io)
       St = write(io->FD, io->Buf, io->BufSize);
       io->Status = St;
 
-      DEBUG_MSG(3, "  IO_write: %s [errno %d] [St %d]\n",
-                g_strerror(errno), errno, St);
+      DEBUG_MSG(3, "  IO_write: %s [errno %d] [St %ld]\n",
+                g_strerror(errno), errno, (glong)St);
 
       if ( St < 0 ) {
          /* Error */
@@ -486,7 +526,7 @@ static gboolean IO_write(IOData_t *io)
          io->BufSize -= St;
          io->Buf = (gchar *)io->Buf + St;
          DataPending = TRUE;
-         DEBUG_MSG(4, "  IO_write: Changing io->Buf (%d)\n", St);
+         DEBUG_MSG(4, "  IO_write: Changing io->Buf (%ld)\n", (glong)St);
 
       } else {
          /* All data in buffer written */
@@ -519,7 +559,7 @@ static gboolean IO_callback(GIOChannel *src, GIOCondition cond, gpointer data)
    gint io_key = GPOINTER_TO_INT(data);
    IOData_t *io = IO_get(io_key);
 
-   // IO_print_cond_status("IO_callback: ", cond, src, io_key);
+   /* IO_print_cond_status("IO_callback: ", cond, src, io_key); */
 
    /* There should be no more glib events on already closed FDs  --Jcid */
    if ( io == NULL ) {
@@ -681,19 +721,29 @@ void a_IO_ccc(int Op, int Branch, int Dir, ChainLink *Info,
          /* Write data using a thread */
          switch (Op) {
          case OpStart:
-            Info->LocalKey = Data1; /* SockFD */
+            {
+            gint *fd = g_new(gint, 1);
+            *fd = *(int*)Data1; /* SockFD */
+            Info->LocalKey = fd;
             break;
+            }
          case OpEnd:
-            a_IO_write_chunk((gint)Info->LocalKey, NULL, 0);
+            _MSG(" Info=%p Info->LocalKey=%p FD=%d Data1=%p\n",
+                 Info, Info->LocalKey, *(int*)Info->LocalKey, Data1);
+            a_IO_write_chunk(*(int*)Info->LocalKey, Info,
+                             NULL, (Data1) ? (size_t)1 : (size_t)0);
+            g_free(Info->LocalKey);
             break;
          case OpSend:
             {
             /* this part submits the data to the thread */
             DataBuf *dbuf = Data1;
-            a_IO_write_chunk((gint)Info->LocalKey, dbuf->Buf, dbuf->Size);
+            a_IO_write_chunk(*(int*)Info->LocalKey, Info,
+                             dbuf->Buf, (size_t)dbuf->Size);
             break;
             }
          case OpAbort:
+            g_free(Info->LocalKey);
             g_free(Info);
             break;
          }
