@@ -36,11 +36,12 @@
 #include <netdb.h>
 
 #include "../msg.h"
-#include "Url.h"
 #include "IO.h"
+#include "Url.h"
 #include "../misc.h"
+#include "../../dpip/dpip.h"
 
-//#define DEBUG_LEVEL 2
+/* #define DEBUG_LEVEL 2 */
 #define DEBUG_LEVEL 4
 #include "../debug.h"
 
@@ -68,7 +69,7 @@ typedef struct {
    gint BufIdx;
    gint BufSize;
 
-   gchar *Tok;
+   gint TokIdx;
    gint TokSize;
    gint TokIsTag;
 
@@ -90,32 +91,15 @@ static GSList *PendingNodes = NULL;
 
 
 /*
- * Task: given a tag and an attribute name, return its value.
- *       (character stuffing is removed here)
- * Return value: the attribute value, or NULL if not present or malformed.
+ * Close a FD handling EINTR
  */
-char *Get_attr_value(char *tag, int tagsize, char *attrname)
+static void Dpi_close_fd(gint fd)
 {
-   char *p, *q, *ltag, quote, *start, *val = NULL;
+   gint st;
 
-   ltag = g_strndup(tag, tagsize);
-   if ((p = strstr(ltag, attrname)) &&
-       (p = strchr(p, '=')) &&
-       (p = strpbrk(p, "'\"")) ) {
-      quote = *p;
-      start = ++p;
-      while ((q = strchr(p, quote)) && q[1] == quote)
-         p = q + 2;
-      if (q) {
-         val = g_strndup(start, q - start);
-         for (p = q = val; (*q = *p); ++p, ++q)
-            if ((*p == '"' || *p == '\'') && p[1] == p[0])
-               ++p;
-      }
-   }
-   g_free(ltag);
-
-   return val;
+   do
+      st = close(fd);
+   while (st < 0 && errno == EINTR);
 }
 
 /*
@@ -126,7 +110,6 @@ static conn_data_t *Dpi_conn_data_new(ChainLink *Info)
    conn_data_t *conn = g_new0(conn_data_t, 1);
 
    conn->Buf = NULL;
-   conn->Tok = NULL;
    conn->InfoRecv = Info;
    return conn;
 }
@@ -147,8 +130,8 @@ static void Dpi_conn_data_free(conn_data_t *conn)
 static void Dpi_append_io_buf(conn_data_t *conn, IOData_t *io)
 {
    if (io->Status > 0) {
-      conn->Buf = g_realloc(conn->Buf, io->Status);
-      memcpy(conn->Buf + conn->BufSize, io->Buf, io->Status);
+      conn->Buf = g_realloc(conn->Buf, conn->BufSize + (gulong)io->Status);
+      memcpy((gchar*)conn->Buf + conn->BufSize, io->Buf, (size_t)io->Status);
       conn->BufSize += io->Status;
       conn->FreeBuf = 0;
    }
@@ -159,10 +142,14 @@ static void Dpi_append_io_buf(conn_data_t *conn, IOData_t *io)
  * Here, a token is either:
  *    a) a dpi tag
  *    b) a raw data chunk
- */
+ *
+ * Return Value: 0 upon a new token, -1 on not enough data.
+ *
+ * TODO: define an API and move this function into libDpip.a.
+*/
 static gint Dpi_get_token(conn_data_t *conn)
 {
-   gint resp = -1;
+   gint i, resp = -1;
    gchar *buf = conn->Buf;
 
    if (conn->FreeBuf || conn->BufIdx == conn->BufSize) {
@@ -174,7 +161,7 @@ static gint Dpi_get_token(conn_data_t *conn)
    }
 
    if (conn->Send2EOF) {
-      conn->Tok = buf + conn->BufIdx;
+      conn->TokIdx = conn->BufIdx;
       conn->TokSize = conn->BufSize - conn->BufIdx;
       conn->BufIdx = conn->BufSize;
       return 0;
@@ -194,7 +181,7 @@ g_free(pbuf);
       if (conn->BufIdx < conn->BufSize) {
          /* found */
          conn->InTag = 1;
-         conn->Tok = buf + conn->BufIdx;
+         conn->TokIdx = conn->BufIdx;
       } else {
          MSG("ERROR: [Dpi_get_token] Can't find token start\n");
          conn->FreeBuf = 1;
@@ -203,13 +190,16 @@ g_free(pbuf);
    }
 
    if (conn->InTag) {
-      /* search for end of tag */
-      while (conn->BufIdx < conn->BufSize && buf[conn->BufIdx] != '>')
-         ++conn->BufIdx;
+      /* search for end of tag (EOT=" '>") */
+      for (i = conn->BufIdx; i < conn->BufSize; ++i)
+         if (buf[i] == '>' && i >= 2 && buf[i-1] == '\'' && buf[i-2] == ' ')
+            break;
+      conn->BufIdx = i;
+
       if (conn->BufIdx < conn->BufSize) {
          /* found EOT */
          conn->TokIsTag = 1;
-         conn->TokSize = buf + conn->BufIdx - conn->Tok + 1;
+         conn->TokSize = conn->BufIdx - conn->TokIdx + 1;
          ++conn->BufIdx;
          conn->InTag = 0;
          resp = 0;
@@ -246,40 +236,46 @@ static void Dpi_parse_token(conn_data_t *conn)
 
    if (conn->Send2EOF) {
       /* we're receiving data chunks from a HTML page */
-      dbuf = a_Chain_dbuf_new(conn->Tok, conn->TokSize, 0);
+      dbuf = a_Chain_dbuf_new(conn->Buf + conn->TokIdx, conn->TokSize, 0);
       a_Chain_fcb(OpSend, conn->InfoRecv, dbuf, "send_page_2eof");
       g_free(dbuf);
       return;
    }
 
-   tag = g_strndup(conn->Tok, conn->TokSize);
-   MSG("Dpi_parse_token: [%s]\n", tag);
-   g_free(tag);
+   tag = g_strndup(conn->Buf + conn->TokIdx, (guint)conn->TokSize);
+   MSG("Dpi_parse_token: {%s}\n", tag);
 
-   cmd = Get_attr_value(conn->Tok, conn->TokSize, "cmd");
+   cmd = a_Dpip_get_attr(conn->Buf + conn->TokIdx, conn->TokSize, "cmd");
    if (strcmp(cmd, "send_status_message") == 0) {
-      msg = Get_attr_value(conn->Tok, conn->TokSize, "msg");
+      msg = a_Dpip_get_attr(conn->Buf + conn->TokIdx, conn->TokSize, "msg");
       a_Chain_fcb(OpSend, conn->InfoRecv, msg, cmd);
       g_free(msg);
 
    } else if (strcmp(cmd, "chat") == 0) {
-      msg = Get_attr_value(conn->Tok, conn->TokSize, "msg");
+      msg = a_Dpip_get_attr(conn->Buf + conn->TokIdx, conn->TokSize, "msg");
       a_Chain_fcb(OpSend, conn->InfoRecv, msg, cmd);
       g_free(msg);
 
+   } else if (strcmp(cmd, "dialog") == 0) {
+      /* For now will send the dpip tag... */
+      a_Chain_fcb(OpSend, conn->InfoRecv, tag, cmd);
+
    } else if (strcmp(cmd, "start_send_page") == 0) {
-      urlstr = Get_attr_value(conn->Tok, conn->TokSize, "url");
+      urlstr = a_Dpip_get_attr(conn->Buf + conn->TokIdx, conn->TokSize, "url");
       a_Chain_fcb(OpSend, conn->InfoRecv, urlstr, cmd);
       g_free(urlstr);
-      /* todo:  Get_attr_value(conn->Tok, conn->TokSize, "send_mode")  */
+      /* todo: a_Dpip_get_attr(conn->Buf + conn->TokIdx,
+       *                       conn->TokSize, "send_mode") */
       conn->Send2EOF = 1;
 
    } else if (strcmp(cmd, "reload_request") == 0) {
-      urlstr = Get_attr_value(conn->Tok, conn->TokSize, "url");
+      urlstr = a_Dpip_get_attr(conn->Buf + conn->TokIdx, conn->TokSize, "url");
       a_Chain_fcb(OpSend, conn->InfoRecv, urlstr, cmd);
       g_free(urlstr);
    }
    g_free(cmd);
+
+   g_free(tag);
 }
 
 /*
@@ -302,7 +298,7 @@ static void Dpi_process_io(int Op, void *Data1, conn_data_t *conn)
    IOData_t *io = Data1;
 
    /* Very useful for debugging: show the data stream as received. */
-   // fwrite(io->Buf, io->Status, 1, stdout);
+   /* fwrite(io->Buf, io->Status, 1, stdout); */
 
    if (Op == IORead) {
       Dpi_append_io_buf(conn, io);
@@ -319,7 +315,7 @@ static void Dpi_process_io(int Op, void *Data1, conn_data_t *conn)
  * Start dpid.
  * Return: 0 starting now, 1 Error.
  */
-static gint Dpi_start_dpid()
+static gint Dpi_start_dpid(void)
 {
    pid_t pid;
    gint st_pipe[2], n, ret = 1;
@@ -333,23 +329,31 @@ static gint Dpi_start_dpid()
    if (pid == 0) {
       /* This is the child process.  Execute the command. */
       gchar *path1 = a_Misc_prepend_user_home(".dillo/dpid");
-      close(st_pipe[0]);
+      Dpi_close_fd(st_pipe[0]);
       if (execl(path1, "dpid", NULL) == -1) {
          g_free(path1);
          if (execlp("dpid", "dpid", NULL) == -1) {
             DEBUG_MSG(4, "Dpi_start_dpid (child): %s\n", g_strerror(errno));
-            write(st_pipe[1], "ERROR", 5);
-            close(st_pipe[1]);
+            do
+               n = write(st_pipe[1], "ERROR", 5);
+            while (n == -1 && errno == EINTR);
+            Dpi_close_fd(st_pipe[1]);
             _exit (EXIT_FAILURE);
          }
       }
    } else if (pid < 0) {
       /* The fork failed.  Report failure.  */
       DEBUG_MSG(4, "Dpi_start_dpid: %s\n", g_strerror(errno));
+      /* close the unused pipe */
+      Dpi_close_fd(st_pipe[0]);
+      Dpi_close_fd(st_pipe[1]);
+
    } else {
       /* This is the parent process, check our child status... */
-      close(st_pipe[1]);
-      n = read(st_pipe[0], buf, 16);
+      Dpi_close_fd(st_pipe[1]);
+      do
+         n = read(st_pipe[0], buf, 16);
+      while (n == -1 && errno == EINTR);
       DEBUG_MSG(2, "Dpi_start_dpid: n = %d\n", n);
       if (n != 5)
          ret = 0;
@@ -364,7 +368,7 @@ static gint Dpi_start_dpid()
  * Make a connection test for a UDS.
  * Return: 0 OK, 1 Not working.
  */
-gint Dpi_check_uds(gchar *uds_name)
+static gint Dpi_check_uds(gchar *uds_name)
 {
    struct sockaddr_un pun;
    gint SockFD, ret = 1;
@@ -374,12 +378,12 @@ gint Dpi_check_uds(gchar *uds_name)
       memset(&pun, 0, sizeof(struct sockaddr_un));
       pun.sun_family = AF_LOCAL;
       strncpy(pun.sun_path, uds_name, sizeof (pun.sun_path));
-   
+
       if ((SockFD = socket(AF_LOCAL, SOCK_STREAM, 0)) == -1 ||
           connect(SockFD, (void*)&pun, D_SUN_LEN(&pun)) == -1) {
          DEBUG_MSG(4, "Dpi_check_uds: %s %s\n", g_strerror(errno), uds_name);
       } else {
-         close(SockFD);
+         Dpi_close_fd(SockFD);
          ret = 0;
       }
    }
@@ -390,13 +394,13 @@ gint Dpi_check_uds(gchar *uds_name)
  * Return the directory where the UDS are in,
  * NULL if it can't be found.
  */
-gchar *Dpi_get_dpid_uds_dir(void)
+static gchar *Dpi_get_dpid_uds_dir(void)
 {
    FILE *in;
    gchar *saved_name_filename;    /*  :)  */
    gchar dpid_uds_dir[256], *p;
 
-   saved_name_filename = 
+   saved_name_filename =
       g_strconcat(g_get_home_dir(), "/", ".dillo/dpi_socket_dir", NULL);
    in = fopen(saved_name_filename, "r");
    g_free(saved_name_filename);
@@ -419,7 +423,7 @@ gchar *Dpi_get_dpid_uds_dir(void)
 /*
  * Return the dpid's UDS name, NULL on failure.
  */
-gchar *Dpi_get_dpid_uds_name(void)
+static gchar *Dpi_get_dpid_uds_name(void)
 {
    gchar *dpid_uds_dir, *dpid_uds_name = NULL;
 
@@ -434,7 +438,7 @@ gchar *Dpi_get_dpid_uds_name(void)
  * Confirm that the dpid is running. If not, start it.
  * Return: 0 running OK, 1 starting (EAGAIN), 2 Error.
  */
-gint Dpi_check_dpid()
+static gint Dpi_check_dpid(void)
 {
    static gint starting = 0;
    gchar *dpid_uds_name;
@@ -479,10 +483,11 @@ gint Dpi_check_dpid()
  *       that connecting each time significantly degrades performance,
  *       an optimized approach can be tried.
  */
-gchar *Dpi_get_server_uds_name(gchar *server_name)
+static gchar *Dpi_get_server_uds_name(const gchar *server_name)
 {
    gchar *dpid_uds_dir, *dpid_uds_name = NULL,
          *server_uds_name = NULL;
+   gint st;
 
    g_return_val_if_fail (server_name != NULL, NULL);
    DEBUG_MSG(2, "Dpi_get_server_uds_name:: server_name = [%s]\n", server_name);
@@ -490,8 +495,9 @@ gchar *Dpi_get_server_uds_name(gchar *server_name)
    dpid_uds_dir = Dpi_get_dpid_uds_dir();
    if (dpid_uds_dir) {
       struct sockaddr_un dpid;
-      gint sock, req_sz, rdlen, buflen;
+      gint sock, req_sz, rdlen;
       gchar buf[128], *cmd, *request, *rply;
+      size_t buflen;
 
       /* Get the server's uds name from dpid */
       sock = socket(AF_LOCAL, SOCK_STREAM, 0);
@@ -499,14 +505,16 @@ gchar *Dpi_get_server_uds_name(gchar *server_name)
       dpid_uds_name = g_strconcat(dpid_uds_dir, "/", "dpid.srs", NULL);
       _MSG("dpid_uds_name = [%s]\n", dpid_uds_name);
       strncpy(dpid.sun_path, dpid_uds_name, sizeof(dpid.sun_path));
-      
+
       if (connect(sock, (struct sockaddr *) &dpid, sizeof(dpid)) == -1)
          perror("connect");
       /* ask dpid to check the server plugin and send its UDS name back */
-      request = g_strdup_printf("<dpi cmd='check_server' msg='%s'>",
-                                server_name);
+      request = a_Dpip_build_cmd("cmd=%s msg=%s", "check_server", server_name);
       DEBUG_MSG(2, "[%s]\n", request);
-      if (write(sock, request, strlen(request)) == -1)
+      do
+         st = write(sock, request, strlen(request));
+      while (st < 0 && errno == EINTR);
+      if (st < 0 && errno != EINTR)
          perror("writing request");
       g_free(request);
       shutdown(sock, 1); /* signals no more writes to dpid */
@@ -515,22 +523,27 @@ gchar *Dpi_get_server_uds_name(gchar *server_name)
       rply = NULL;
       buf[0] = '\0';
       buflen = sizeof(buf)/sizeof(buf[0]);
-      for (req_sz = 0;
-            (rdlen = read(sock, buf, buflen)) != 0 && rdlen != -1;
-            req_sz += rdlen) {
-         rply = (char *) realloc(rply, req_sz + rdlen + 1);
+      for (req_sz = 0; (rdlen = read(sock, buf, buflen)) != 0;
+           req_sz += rdlen) {
+         if (rdlen == -1 && errno == EINTR)
+               continue;
+         if (rdlen == -1) {
+            perror(" ** Dpi_get_server_uds_name **");
+            break;
+         }
+         rply = g_realloc(rply, (guint)(req_sz + rdlen + 1));
          if (req_sz == 0)
             rply[0] = '\0';
-         strncat(rply, buf, rdlen);
+         strncat(rply, buf, (size_t)rdlen);
       }
-      close(sock);
+      Dpi_close_fd(sock);
       DEBUG_MSG(2, "rply = [%s]\n", rply);
 
       /* Parse reply */
-      if (rply) {
-         cmd = Get_attr_value(rply, strlen(rply), "cmd");
+      if (rdlen == 0 && rply) {
+         cmd = a_Dpip_get_attr(rply, (gint)strlen(rply), "cmd");
          if (strcmp(cmd, "send_data") == 0)
-            server_uds_name = Get_attr_value(rply, strlen(rply), "msg");
+            server_uds_name = a_Dpip_get_attr(rply, (gint)strlen(rply), "msg");
          g_free(cmd);
          g_free(rply);
       }
@@ -547,7 +560,7 @@ gchar *Dpi_get_server_uds_name(gchar *server_name)
  * We have to ask 'dpid' (dpi daemon) for the UDS of the target dpi server.
  * Once we have it, then the proper file descriptor is returned (-1 on error).
  */
-static gint Dpi_connect_socket(gchar *server_name, gint retry)
+static gint Dpi_connect_socket(const gchar *server_name, gint retry)
 {
    char *server_uds_name;
    struct sockaddr_un pun;
@@ -619,6 +632,16 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             a_Url_free(Info->LocalKey);
             a_Chain_fcb(OpEnd, Info, NULL, NULL);
             break;
+         case OpAbort:
+            list = g_slist_find_custom(
+                      PendingNodes, Info->LocalKey, Dpi_node_cmp);
+            if (list) {
+               /* The connection is not pending anymore */
+               PendingNodes = g_slist_remove(PendingNodes, list->data);
+            }
+            a_Url_free(Info->LocalKey);
+            a_Chain_fcb(OpAbort, Info, NULL, NULL);
+            break;
          }
       }
 
@@ -630,11 +653,13 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             if ((st = Dpi_check_dpid()) == 0) {
                SockFD = Dpi_connect_socket(Data1, TRUE);
                if (SockFD != -1) {
-                  Info->LocalKey = (void *)SockFD;
+                  gint *fd = g_new(gint, 1);
+                  *fd = SockFD;
+                  Info->LocalKey = fd;
                   a_Chain_link_new(Info, a_Dpi_ccc, BCK, a_IO_ccc, 3, 2);
-                  a_Chain_bcb(OpStart, Info, (void *)SockFD, NULL);
+                  a_Chain_bcb(OpStart, Info, Info->LocalKey, NULL);
                   /* tell the capi to start the receiving branch */
-                  a_Chain_fcb(OpSend, Info, (void *)SockFD, "SockFD");
+                  a_Chain_fcb(OpSend, Info, Info->LocalKey, "SockFD");
                }
             }
 
@@ -652,10 +677,13 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             break;
          case OpEnd:
             a_Chain_bcb(OpEnd, Info, NULL, NULL);
+            g_free(Info->LocalKey);
             g_free(Info);
             break;
          case OpAbort:
             MSG("a_Dpi_ccc: OpAbort[2B], Not implemented\n");
+            g_free(Info->LocalKey);
+            g_free(Info);
             break;
          }
       } else {  /* FWD */
@@ -696,7 +724,7 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             {
             IOData_t *io2;
             Info->LocalKey = Dpi_conn_data_new(Info);
-            io2 = a_IO_new(IORead, (gint)Data1); /* SockFD */
+            io2 = a_IO_new(IORead, *(int*)Data1); /* SockFD */
             a_IO_set_buf(io2, NULL, IOBufLen);
             a_Chain_link_new(Info, a_Dpi_ccc, BCK, a_IO_ccc, 2, 3);
             a_Chain_bcb(OpStart, Info, io2, NULL);
@@ -716,6 +744,16 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
                a_Dpi_ccc(OpEnd, 1, FWD, P_Info, NULL, NULL);
             }
             break;
+         case OpStop:
+            /* Stop transfer, abort the pending node.
+             * (Data1 = Url, Data2 = NULL) */
+            list = g_slist_find_custom(PendingNodes, Data1, Dpi_node_cmp);
+            if (list) {
+               ChainLink *P_Info = list->data;
+               /* Abort this requesting CCC */
+               a_Dpi_ccc(OpAbort, 1, FWD, P_Info, Data1, NULL);
+            }
+            break;
          case OpAbort:
             MSG(" Not implemented\n");
             break;
@@ -734,7 +772,7 @@ void a_Dpi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
  */
 void a_Dpi_bye_dpid()
 {
-   char *DpiBye_cmd = "<dpi cmd='DpiBye'>";
+   char *DpiBye_cmd;
    struct sockaddr_un sa;
    size_t sun_path_len, addr_len;
    char *srs_name;
@@ -753,8 +791,54 @@ void a_Dpi_bye_dpid()
    if (connect(new_socket, (struct sockaddr *) &sa, addr_len) == -1) {
       DEBUG_MSG(4, "a_Dpi_bye_dpid: %s\n", g_strerror(errno));
       fprintf(stderr, "%s\n", sa.sun_path);
-   } 
+   }
+   DpiBye_cmd = a_Dpip_build_cmd("cmd=%s", "DpiBye");
    (void) write(new_socket, DpiBye_cmd, strlen(DpiBye_cmd));
-   close(new_socket);
+   g_free(DpiBye_cmd);
+   Dpi_close_fd(new_socket);
+}
+
+
+/*
+ * Send a command to a dpi server, and block until the answer is got.
+ * Return value: the dpip tag answer as an string, NULL on error.
+ */
+char *a_Dpi_send_blocking_cmd(const gchar *server_name, const gchar *cmd)
+{
+   gint i, cst, SockFD;
+   ssize_t st;
+   gchar buf[16384], *retval = NULL;
+
+   /* test the dpid, and wait a bit for it to start if necessary */
+   for (i = 0; (cst = Dpi_check_dpid()) == 1 && i < 2; ++i)
+      sleep(1);
+
+   if (cst != 0)
+      return retval;
+
+   SockFD = Dpi_connect_socket(server_name, TRUE);
+   if (SockFD != -1) {
+      /* todo: handle the case of (st < strlen(cmd)) */
+      do
+         st = write(SockFD, cmd, strlen(cmd));
+      while (st == -1 && errno == EINTR);
+
+      /* todo: if the answer is too long... */
+      do
+         st = read(SockFD, buf, 16384);
+      while (st < 0 && errno == EINTR);
+
+      if (st == -1)
+         perror("[a_Dpi_send_blocking_cmd]");
+      else if (st > 0)
+         retval = g_strndup(buf, (guint)st);
+
+      Dpi_close_fd(SockFD);
+
+   } else {
+      perror("[a_Dpi_send_blocking_cmd]");
+   }
+
+   return retval;
 }
 
