@@ -21,6 +21,8 @@
 #include "capi.h"
 #include "nav.h"
 #include "misc.h"
+#include "bookmark.h"  /* for prototypes */
+#include "../dpip/dpip.h"
 
 
 
@@ -39,11 +41,15 @@ void a_Bookmarks_init(void)
  */
 void a_Bookmarks_chat_add(BrowserWindow *Bw, char *Cmd, char *answer)
 {
-   char *cmd1 = "<dpi cmd='chat' msg='Hi server'>";
-   char *cmd2 = "<dpi cmd='chat' msg='I want to set a bookmark'>";
-   char *cmd3 = "<dpi cmd='chat' msg='Sure it is!'>";
+   static char *cmd1 = NULL, *cmd2 = NULL, *cmd3 = NULL, *cmd4 = NULL;
    static BrowserWindow *bw = NULL;
-   static gchar *cmd4 = NULL;
+
+   if (!cmd1) {
+      cmd1 = a_Dpip_build_cmd("cmd=%s msg=%s", "chat", "Hi server");
+      cmd2 = a_Dpip_build_cmd("cmd=%s msg=%s", "chat",
+                              "I want to set a bookmark");
+      cmd3 = a_Dpip_build_cmd("cmd=%s msg=%s", "chat", "Sure it is!");
+   }
 
    _MSG("a_Bookmarks_chat_add\n answer=%s\n", answer ? answer : "(null)");
 
@@ -80,19 +86,18 @@ void a_Bookmarks_chat_add(BrowserWindow *Bw, char *Cmd, char *answer)
 void a_Bookmarks_add(GtkWidget *widget, gpointer client_data)
 {
    BrowserWindow *bw = (BrowserWindow *)client_data;
-   gchar *title, *t;
    DilloUrl *url;
-   gchar *cmd = NULL;
+   const gchar *title;
+   gchar *cmd;
 
    url = a_Menu_popup_get_url(bw);
    g_return_if_fail(url != NULL);
 
    /* if the page has no title, we'll use the url string */
-   title = (gchar *) a_History_get_title_by_url(url, 1);
-   title = (t = a_Misc_stuff_chars(title, "'\"")) ? t : title;
+   title = a_History_get_title_by_url(url, 1);
 
-   cmd = g_strdup_printf("<dpi cmd='add_bookmark' url='%s' title='%s'>",
-                         URL_STR(url), title);
+   cmd = a_Dpip_build_cmd("cmd=%s url=%s title=%s",
+                          "add_bookmark", URL_STR(url), title);
    a_Bookmarks_chat_add(bw, cmd, NULL);
    g_free(cmd);
 }
