@@ -16,6 +16,7 @@
  */
 
 #include <string.h>
+#include <unistd.h>   /* for pipe */
 
 #include "msg.h"
 #include "capi.h"
@@ -25,6 +26,9 @@
 #include "interface.h"
 #include "history.h"
 #include "nav.h"
+#include "misc.h"
+#include "dpiapi.h"
+#include "../dpip/dpip.h"
 
 /* for testing dpi chat */
 #include "bookmark.h"
@@ -33,7 +37,8 @@
 #include "debug.h"
 
 typedef struct {
-   DilloWeb *web;        /* Necessary for <dpi cmd='open_url' ...> */
+   DilloWeb *web;
+   DilloUrl *url; /* local copy of web->url. Used when the latter is freed */
    void *bw;
    gchar *server;
    gchar *datastr;
@@ -50,7 +55,8 @@ typedef struct {
 /* Flags for conn */
 enum {
    PENDING = 1,
-   TIMEOUT = 2   /* unused */
+   TIMEOUT = 2,  /* unused */
+   ABORTED = 4
 };
 
 /*
@@ -83,6 +89,7 @@ static dpi_conn_t *
 
    conn = g_new(dpi_conn_t, 1);
    conn->web = web;
+   conn->url = (web) ? a_Url_dup(web->url) : NULL;
    conn->bw = bw;
    conn->server = g_strdup(server);
    conn->datastr = g_strdup(datastr);
@@ -123,6 +130,8 @@ static void Capi_dpi_conn_unref(dpi_conn_t *conn)
             for (j = i; j + 1 < DpiConnSize; ++j)
                DpiConn[j] = DpiConn[j + 1];
             --DpiConnSize;
+            /* free dynamic memory */
+            a_Url_free(conn->url);
             g_free(conn->server);
             g_free(conn->datastr);
             g_free(conn);
@@ -156,7 +165,7 @@ static void Capi_dpi_conn_resume(void)
    for (i = 0; i < DpiConnSize; ++i)
       if (DpiConn[i]->Flags & PENDING) {
          dpi_conn_t *conn = DpiConn[i];
-         dbuf = a_Chain_dbuf_new(conn->datastr, strlen(conn->datastr), 0);
+         dbuf = a_Chain_dbuf_new(conn->datastr,(gint)strlen(conn->datastr), 0);
          a_Capi_ccc(OpSend, 1, BCK, conn->InfoSend, dbuf, NULL);
          g_free(dbuf);
          conn->Flags &= ~PENDING;
@@ -188,15 +197,15 @@ static gint Capi_dpi_conn_timeout(gpointer data)
    dpi_conn_timeout_id = 0;
    return FALSE;
 }
-   
+
 /* ------------------------------------------------------------------------- */
 
 /*
  * Safety test: only allow dpi-urls from dpi-generated pages.
  */
-gint Capi_verify_dpi_url_request(DilloWeb *web)
+static gint Capi_verify_dpi_url_request(DilloWeb *web)
 {
-   DilloUrl *referer = NULL;
+   DilloUrl *referer;
    gint allow = FALSE;
 
    /* test POST and GET */
@@ -229,8 +238,11 @@ gint a_Capi_url_uses_dpi(gchar *url_str, gchar **server_ptr)
 
    if (g_strncasecmp(url_str, "dpi:/", 5) == 0) {
       /* dpi prefix, get this server's name */
-      p = strchr(url_str + 5, '/');
-      server = p ? g_strndup(url_str + 5, p - url_str - 5) : g_strdup("?");
+      if ((p = strchr(url_str + 5, '/')) != NULL)
+         server = g_strndup(url_str + 5, (guint)(p - url_str - 5));
+      else
+         server = g_strdup("?");
+
       if (strcmp(server, "bm") == 0) {
          g_free(server);
          server = g_strdup("bookmarks");
@@ -241,11 +253,37 @@ gint a_Capi_url_uses_dpi(gchar *url_str, gchar **server_ptr)
 
    } else if (g_strncasecmp(url_str, "https:/", 7) == 0) {
       server = g_strdup("https");
+   } else if (g_strncasecmp(url_str, "file:", 5) == 0) {
+      server = g_strdup("file");
+   } else if (g_strncasecmp(url_str, "data:", 5) == 0) {
+      server = g_strdup("datauri");
    }
 
    return ((*server_ptr = server) ? 1 : 0);
 }
 
+/*
+ * Build the dpip command tag, according to URL and server.
+ * todo: make it PROXY-aware (AFAIS, it should be easy)
+ */
+static gchar *Capi_dpi_build_cmd(DilloUrl *url, gchar *server)
+{
+   gchar *cmd, *http_query;
+
+   if (strcmp(server, "https") == 0) {
+      /* Let's be kind and make the HTTP query string for the dpi */
+      http_query = a_Http_make_query_str(url, FALSE);
+      cmd = a_Dpip_build_cmd("cmd=%s url=%s query=%s",
+                             "open_url", URL_STR(url), http_query);
+      g_free(http_query);
+
+   } else {
+      /* For everyone else, the url_str is enough... */
+      cmd = a_Dpip_build_cmd("cmd=%s url=%s", "open_url", URL_STR(url));
+   }
+   return cmd;
+}
+
 /*
  * Most used function.
  * todo: clean up the ad-hoc bindings with an API that allows dynamic
@@ -262,7 +300,7 @@ gint a_Capi_open_url(DilloWeb *web, CA_Callback_t Call, void *CbData)
 
    if (a_Capi_url_uses_dpi(url_str, &server) && !Call) {
 
-      MSG("  url_str = %s\n", url_str);
+      _MSG("  url_str = %s\n", url_str);
 
       if (g_strncasecmp(url_str, "dpi:/", 5) == 0) {
          /* safety check... */
@@ -282,7 +320,7 @@ gint a_Capi_open_url(DilloWeb *web, CA_Callback_t Call, void *CbData)
 
       if (safe && reload) {
          /* Send dpip command */
-         cmd = g_strdup_printf("<dpi cmd='open_url' url='%s'>", url_str);
+         cmd = Capi_dpi_build_cmd(web->url, server);
          a_Capi_dpi_send_cmd(web, web->bw, cmd, server, 1);
          g_free(cmd);
 
@@ -325,15 +363,11 @@ gint a_Capi_dpi_send_cmd(DilloWeb *web, void *bw, char *cmd, char *server,
    dpi_conn_t *conn;
    DataBuf *dbuf;
 
-   //**/sleep(1);
-   //**/a_Interface_msg(bw, cmd);
-
    if (flags & 1) {
       /* open a new connection to server */
 
       /* Create a new connection data struct and add it to the list */
       conn = Capi_dpi_conn_new(web, bw, server, cmd);
-      Capi_dpi_conn_ref(conn);
       /* start the CCC operations */
       a_Capi_ccc(OpStart, 1, BCK, conn->InfoSend, conn, server);
 
@@ -342,7 +376,7 @@ gint a_Capi_dpi_send_cmd(DilloWeb *web, void *bw, char *cmd, char *server,
       conn = Capi_dpi_conn_find(server);
       if (conn) {
          /* found */
-         dbuf = a_Chain_dbuf_new(cmd, strlen(cmd), 0);
+         dbuf = a_Chain_dbuf_new(cmd, (gint)strlen(cmd), 0);
          a_Capi_ccc(OpSend, 1, BCK, conn->InfoSend, dbuf, NULL);
          g_free(dbuf);
       } else {
@@ -369,6 +403,8 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
          /* Command sending branch */
          switch (Op) {
          case OpStart:
+            conn = Data1;
+            Capi_dpi_conn_ref(conn);
             Info->LocalKey = Data1;
             a_Chain_link_new(Info, a_Capi_ccc, BCK, a_Dpi_ccc, 2, 1);
             a_Chain_bcb(OpStart, Info, Data2, NULL);
@@ -395,7 +431,7 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             } else if (strcmp(Data2, "SockFD") == 0) {
                /* start the receiving branch */
                dpi_conn_t *conn = Info->LocalKey;
-               conn->SockFD = (gint)Data1;
+               conn->SockFD = *(int*)Data1;
                a_Capi_ccc(OpStart, 2, BCK, a_Chain_new(),Info->LocalKey, NULL);
             } else if (strcmp(Data2, "DpidOK") == 0) {
                /* send the data inmediatly! */
@@ -403,7 +439,7 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             } else if (strcmp(Data2, "DpidEAGAIN") == 0) {
                /* set a timeout function to retry later... */
                if (dpi_conn_timeout_id == 0)
-                  dpi_conn_timeout_id = 
+                  dpi_conn_timeout_id =
                      gtk_timeout_add(250, (GtkFunction) Capi_dpi_conn_timeout,
                                      NULL);
             }
@@ -426,11 +462,11 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
          case OpStart:
             {
             dpi_conn_t *conn = Data1;
-            Info->LocalKey = Data1;
             Capi_dpi_conn_ref(conn);
+            Info->LocalKey = Data1;
             conn->InfoRecv = Info;
             a_Chain_link_new(Info, a_Capi_ccc, BCK, a_Dpi_ccc, 3, 2);
-            a_Chain_bcb(OpStart, Info, (void *)conn->SockFD, "SockFD");
+            a_Chain_bcb(OpStart, Info, &conn->SockFD, "SockFD");
             break;
             }
          case OpStop:
@@ -444,17 +480,28 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
          switch (Op) {
          case OpSend:
             conn = Info->LocalKey;
-            if (strcmp(Data2, "send_status_message") == 0) {
-               a_Interface_msg(conn->bw, Data1);
+            if (conn->Flags & ABORTED ||
+                (conn->web && !a_Web_valid(conn->web))) {
+               /* there's no client for this transfer!*/
+               _MSG("  ** Capi 2F catched an invalid 'web' structure\n");
+               /* this flag is used just in case the same memory address
+                * is reused for a new 'web' and the test passes */
+               conn->Flags &= ABORTED;
+               /* Make the dpi module stop this transfer */
+               a_Chain_bcb(OpStop, Info, conn->url, NULL);
+
+            } else if (strcmp(Data2, "send_status_message") == 0) {
+               a_Interface_msg(conn->bw, "%s", Data1);
             } else if (strcmp(Data2, "chat") == 0) {
-               a_Interface_msg(conn->bw, Data1);
+               a_Interface_msg(conn->bw, "%s", Data1);
                a_Bookmarks_chat_add(NULL, NULL, Data1);
+            } else if (strcmp(Data2, "dialog") == 0) {
+               a_Dpiapi_dialog(conn->bw, conn->server, Data1);
             } else if (strcmp(Data2, "start_send_page") == 0) {
                /* start the pipe-writing chain */
                a_Capi_ccc(OpStart, 3, BCK, a_Chain_new(), Info->LocalKey,NULL);
                /* let the dpi know the reading end of the pipe */
-               a_Chain_bcb(OpSend, Info, (void *)conn->DpiPipe[0],
-                           conn->web->url);
+               a_Chain_bcb(OpSend, Info, &conn->DpiPipe[0], conn->web->url);
             } else if (strcmp(Data2, "send_page_2eof") == 0) {
                a_Capi_ccc(OpSend, 3, BCK, conn->InfoPipe, Data1, NULL);
             } else if (strcmp(Data2, "reload_request") == 0) {
@@ -466,7 +513,6 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
             dpi_conn_t *conn = Info->LocalKey;
             a_Chain_del_link(Info, BCK);
             conn->InfoRecv = NULL;
-            Capi_dpi_conn_unref(conn);
             if (conn->InfoSend) {
                /* Propagate OpEnd to the sending branch too */
                a_Capi_ccc(OpEnd, 1, BCK, conn->InfoSend, NULL, NULL);
@@ -475,6 +521,7 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
                /* Propagate OpEnd to the pipe branch too */
                a_Capi_ccc(OpEnd, 3, BCK, conn->InfoPipe, NULL, NULL);
             }
+            Capi_dpi_conn_unref(conn);
             g_free(Info);
             break;
             }
@@ -500,14 +547,14 @@ void a_Capi_ccc(int Op, int Branch, int Dir, ChainLink *Info,
                return;
             }
             a_Chain_link_new(Info, a_Capi_ccc, BCK, a_IO_ccc, 3, 3);
-            a_Chain_bcb(OpStart, Info, (void *)conn->DpiPipe[1], "SockFD");
+            a_Chain_bcb(OpStart, Info, &conn->DpiPipe[1], "SockFD");
             break;
             }
          case OpSend:
             a_Chain_bcb(OpSend, Info, Data1, NULL);
             break;
          case OpEnd:
-            a_Chain_bcb(OpEnd, Info, NULL, NULL);
+            a_Chain_bcb(OpEnd, Info, (void*)1, NULL);
             Capi_dpi_conn_unref(Info->LocalKey);
             g_free(Info);
             break;
