@@ -377,7 +377,7 @@ DwIterator* a_Dw_widget_iterator   (DwWidget *widget,
 
 /* for convenience */
 #define     a_Dw_iterator_unhighlight(it, l) ((it) ? \
-                                    (it)->highlight(it, -1, -1, l) : FALSE)
+                                    (it)->highlight(it, -1, -1, l) : (void)0)
 void        a_Dw_iterator_scroll_to  (DwIterator *it1,
                                       DwIterator *it2,
                                       gint start,
