@@ -34,6 +34,9 @@
 #include "interface.h"
 #include "menu.h"
 #include "selection.h"
+#include "gtk_ext_menu.h"
+#include "gtk_ext_menu_item.h"
+#include "gtk_menu_title.h"
 
 /*
  * Forward declarations
@@ -69,7 +72,7 @@ static GtkWidget* Menu_new(GtkWidget *menubar, const char *name,
 /*
  * Pick the right URL for this popup.
  */
-void Menu_pick_url (GtkWidget *widget, gpointer client_data)
+static void Menu_pick_url (GtkWidget *widget, gpointer client_data)
 {
    BrowserWindow *bw = (BrowserWindow *)client_data;
 
@@ -117,6 +120,19 @@ static GtkWidget *
    return menuitem;
 }
 
+/*
+ * Add a menu title.
+ */
+static GtkWidget *Menu_add_title (GtkWidget *menu, const char *name)
+{
+   GtkWidget *menuitem;
+
+   menuitem = a_Gtk_menu_title_new(name);
+   gtk_menu_append(GTK_MENU(menu), menuitem);
+   gtk_widget_show(menuitem);
+   return menuitem;
+}
+
 /*
  * Add a separator into the menu.
  */
@@ -179,7 +195,7 @@ GtkWidget *a_Menu_popup_op_new(BrowserWindow *bw)
 
    menu = gtk_menu_new();
    Menu_sep(menu);
-   Menu_add(menu, "       PAGE OPTIONS", NULL, bw, NULL, NULL);
+   Menu_add_title(menu, "PAGE OPTIONS");
    Menu_sep(menu);
    Menu_add(menu, "View page Source", NULL, bw,
             a_Commands_viewsource_callback, bw);
@@ -203,7 +219,7 @@ GtkWidget *a_Menu_popup_op_new(BrowserWindow *bw)
 /*
  * Set the popup's primary DilloUrl
  */
-void a_Menu_popup_set_url(BrowserWindow *bw, DilloUrl *url)
+void a_Menu_popup_set_url(BrowserWindow *bw, const DilloUrl *url)
 {
    if (bw->menu_popup.url)
       a_Url_free(bw->menu_popup.url);
@@ -216,13 +232,23 @@ void a_Menu_popup_set_url(BrowserWindow *bw, DilloUrl *url)
  * (If the need arises, this can be extended using an URL array in the popup
  * structure, and by setting the to-be-picked-URL-index in the popup)
  */
-void a_Menu_popup_set_url2(BrowserWindow *bw, DilloUrl *url)
+void a_Menu_popup_set_url2(BrowserWindow *bw, const DilloUrl *url)
 {
    if (bw->menu_popup.url2)
       a_Url_free(bw->menu_popup.url2);
    bw->menu_popup.url2 = a_Url_dup(url);
 }
 
+/*
+ * Remove the popup's secondary DilloUrl.
+ * (This is done at popup time, so and old "url2" can't linger)
+ */
+void a_Menu_popup_clear_url2(GtkWidget *menu_popup)
+{
+   /* clear the url2 tag (currently "over_image" is the only one using it) */
+   gtk_object_remove_data(GTK_OBJECT (menu_popup), "url2");
+}
+
 /*
  * Get the popup's DilloUrl
  */
@@ -267,13 +293,13 @@ GtkWidget *a_Menu_popup_history_new(BrowserWindow *bw, gint direction)
    GtkWidget *menu, *menu_item;
    GString *text = g_string_sized_new(128);
 
-   menu = gtk_menu_new();
+   menu = a_Gtk_ext_menu_new();
    Menu_sep(menu);
    if (direction>0) {
-      Menu_add(menu, "     FOLLOWING PAGES     ", NULL, bw, NULL, NULL);
+      Menu_add_title(menu, "FOLLOWING PAGES");
       i = a_Nav_stack_ptr(bw) + 1;
    } else {
-      Menu_add(menu, "     PREVIOUS PAGES     ", NULL, bw, NULL, NULL);
+      Menu_add_title(menu, "PREVIOUS PAGES");
       i = a_Nav_stack_ptr(bw) - 1;
    }
    Menu_sep(menu);
@@ -288,7 +314,9 @@ GtkWidget *a_Menu_popup_history_new(BrowserWindow *bw, gint direction)
          g_string_truncate(text, 64);
          g_string_append(text, "...");
       }
-      menu_item = Menu_add(menu, text->str, NULL, bw, NULL, NULL);
+      menu_item = a_Gtk_ext_menu_item_new_with_label(text->str);
+      gtk_menu_append(GTK_MENU(menu), menu_item);
+      gtk_widget_show(menu_item);
       /* attach the nav_stack index to the menu item */
       gtk_object_set_data(GTK_OBJECT (menu_item), "nav_idx",
                           GINT_TO_POINTER(i));
@@ -302,9 +330,19 @@ GtkWidget *a_Menu_popup_history_new(BrowserWindow *bw, gint direction)
       gtk_signal_connect (
          GTK_OBJECT (menu_item), "deselect",
          GTK_SIGNAL_FUNC (Menu_popup_history_deselect_callback), bw);
+
       gtk_signal_connect (
-         GTK_OBJECT (menu_item), "button-press-event",
-         GTK_SIGNAL_FUNC (a_Commands_historypress_callback), bw);
+         GTK_OBJECT (menu_item), "activate",
+         GTK_SIGNAL_FUNC (a_Commands_history_callback_same_bw), bw);
+      gtk_signal_connect (
+         GTK_OBJECT (menu_item), "activate1",
+         GTK_SIGNAL_FUNC (a_Commands_history_callback_same_bw), bw);
+      gtk_signal_connect (
+         GTK_OBJECT (menu_item), "activate2",
+         GTK_SIGNAL_FUNC (a_Commands_history_callback_new_bw), bw);
+      gtk_signal_connect (
+         GTK_OBJECT (menu_item), "activate3",
+         GTK_SIGNAL_FUNC (a_Commands_history_callback_same_bw), bw);
    }
 
    g_string_free(text, TRUE);
@@ -322,7 +360,7 @@ GtkWidget *a_Menu_popup_ol_new(BrowserWindow *bw)
 
    menu = gtk_menu_new();
    Menu_sep(menu);
-   Menu_add(menu, "       LINK OPTIONS", NULL, bw, NULL, NULL);
+   Menu_add_title(menu, "LINK OPTIONS");
    Menu_sep(menu);
    Menu_add(menu, "Open Link in New Window", NULL, bw,
             a_Commands_open_link_nw_callback, bw);
@@ -356,7 +394,7 @@ GtkWidget *a_Menu_popup_oi_new (BrowserWindow *bw)
 
    menu = gtk_menu_new();
    Menu_sep(menu);
-   Menu_add(menu, "       IMAGE OPTIONS", NULL, bw, NULL, NULL);
+   Menu_add_title(menu, "IMAGE OPTIONS");
    Menu_sep(menu);
    Menu_add(menu, "Isolate Image", NULL, bw,
             a_Commands_open_link_callback, bw);
@@ -383,7 +421,7 @@ GtkWidget *a_Menu_popup_ob_new (BrowserWindow *bw)
 
    menu = gtk_menu_new();
    Menu_sep(menu);
-   Menu_add(menu, "    BUG METER OPTIONS", NULL, bw, NULL, NULL);
+   Menu_add_title(menu, "BUG METER OPTIONS");
    Menu_sep(menu);
    Menu_add(menu, "Validate URL with W3C", NULL, bw,
             a_Commands_ob_w3c_callback, bw);
