@@ -10,7 +10,7 @@
  * (at your option) any later version.
  */
 
-#include <gtk/gtk.h>
+#include <glib.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -66,29 +66,6 @@ gchar *a_Misc_escape_chars(const gchar *str, gchar *esc_set)
    return p;
 }
 
-/*
- * Use character stuffing for characters given in 'esc_set'.
- * Return value: New string, or NULL if there's no need to stuff.
- */
-gchar *a_Misc_stuff_chars(const gchar *str, const gchar *esc_set)
-{
-   gint i;
-   gchar *p, *sstr = NULL;
-
-   if ((p = strpbrk(str, esc_set))) {
-      /* start stuffing */
-      GString *gstr = g_string_new(str);
-      for (i = p - str; gstr->str[i]; ++i)
-         if (strchr(esc_set, gstr->str[i])) {
-            g_string_insert_c(gstr, i, gstr->str[i]);
-            ++i;
-         }
-      sstr = gstr->str;
-      g_string_free(gstr, FALSE);
-   }
-   return sstr;
-}
-
 /*
  * Case insensitive strstr
  */
@@ -216,56 +193,131 @@ gint *a_Misc_strsplitposdup(gint *pos)
    return pos2;
 }
 
+/* TODO: could use dStr ADT! */
+typedef struct ContentType_ {
+   const char *str;
+   int len;
+} ContentType_t;
+
+static const ContentType_t MimeTypes[] = {
+   { "application/octet-stream", 24 },
+   { "text/html", 9 },
+   { "text/plain", 10 },
+   { "image/gif", 9 },
+   { "image/png", 9 },
+   { "image/jpeg", 10 },
+   { NULL, 0 }
+};
+
 /*
- * Detects 'Content-Type' when the server does not supply one.
+ * Detects 'Content-Type' from a data stream sample.
+ *
  * It uses the magic(5) logic from file(1). Currently, it
  * only checks the few mime types that Dillo supports.
  *
  * 'Data' is a pointer to the first bytes of the raw data.
+ *
+ * Return value: (0 on success, 1 on doubt, 2 on lack of data).
  */
-const gchar *a_Misc_get_content_type_from_data(void *Data, size_t Size)
+int a_Misc_get_content_type_from_data(void *Data, size_t Size, const char **PT)
 {
-   static const gchar *Types[] = {
-      "application/octet-stream",
-      "text/html", "text/plain",
-      "image/gif", "image/png", "image/jpeg",
-   };
-   gint Type = 0;
-   gchar *p = Data;
-   size_t i;
+   int st = 1;      /* default to "doubt' */
+   int Type = 0;    /* default to "application/octet-stream" */
+   char *p = Data;
+   size_t i, non_ascci;
 
    /* HTML try */
    for (i = 0; i < Size && isspace(p[i]); ++i);
-   if ((Size - i >= 5 && !g_strncasecmp(p+i, "<html", 5)) ||
-       (Size - i >= 5 && !g_strncasecmp(p+i, "<head", 5)) ||
-       (Size - i >= 6 && !g_strncasecmp(p+i, "<title", 6)) ||
-       (Size - i >= 14 && !g_strncasecmp(p+i, "<!doctype html", 14))) {
-      Type = 1;
+   if ((Size - i >= 5  && !g_strncasecmp(p+i, "<html", 5)) ||
+       (Size - i >= 5  && !g_strncasecmp(p+i, "<head", 5)) ||
+       (Size - i >= 6  && !g_strncasecmp(p+i, "<title", 6)) ||
+       (Size - i >= 14 && !g_strncasecmp(p+i, "<!doctype html", 14)) ||
+       /* this line is workaround for FTP through the Squid proxy */
+       (Size - i >= 17 && !g_strncasecmp(p+i, "<!-- HTML listing", 17))) {
 
+      Type = 1;
+      st = 0;
    /* Images */
    } else if (Size >= 4 && !g_strncasecmp(p, "GIF8", 4)) {
       Type = 3;
+      st = 0;
    } else if (Size >= 4 && !g_strncasecmp(p, "\x89PNG", 4)) {
       Type = 4;
+      st = 0;
    } else if (Size >= 2 && !g_strncasecmp(p, "\xff\xd8", 2)) {
       /* JPEG has the first 2 bytes set to 0xffd8 in BigEndian - looking
        * at the character representation should be machine independent. */
       Type = 5;
+      st = 0;
 
    /* Text */
    } else {
-      /* We'll assume ASCII if chars are below 128 (after all, this
-       * is a last resort when the server doesn't send Content-Type) */
+      /* We'll assume "text/plain" if the set of chars above 127 is <= 10
+       * in a 256-bytes sample.  Better heuristics are welcomed! :-) */
+      non_ascci = 0;
       Size = MIN (Size, 256);
       for (i = 0; i < Size; i++)
-         if ((guchar) p[i] > 127)
-            break;
-      Type = (i < 12 || i < Size) ? 0 : 2;
+         if ((unsigned char) p[i] > 127)
+            ++non_ascci;
+      if (Size == 256) {
+         Type = (non_ascci > 10) ? 0 : 2;
+         st = 0;
+      } else {
+         Type = (non_ascci > 0) ? 0 : 2;
+      }
    }
 
-   return (Types[Type]);
+   *PT = MimeTypes[Type].str;
+   return st;
 }
 
+/*
+ * Check the server-supplied 'Content-Type' against our detected type.
+ * (some servers seem to default to "text/plain").
+ *
+ * Return value:
+ *  0,  if they match
+ *  -1, if a mismatch is detected
+ *
+ * There're many MIME types Dillo doesn't know, they're handled
+ * as "application/octet-stream" (as the SPEC says).
+ *
+ * A mismatch happens when receiving a binary stream as
+ * "text/plain" or "text/html", or an image that's not an image of its kind.
+ *
+ * Note: this is a basic security procedure.
+ *
+ */
+int a_Misc_content_type_check(const char *EntryType, const char *DetectedType)
+{
+   int i;
+   int st = -1;
+
+   MSG("Type check:  [Srv: %s  Det: %s]\n", EntryType, DetectedType);
+
+   if (!EntryType)
+      return 0; /* there's no mismatch without server type */
+
+   for (i = 1; MimeTypes[i].str; ++i)
+      if (g_strncasecmp(EntryType, MimeTypes[i].str, MimeTypes[i].len) == 0)
+         break;
+
+   if (!MimeTypes[i].str) {
+      /* type not found, no mismatch */
+      st = 0;
+   } else if (g_strncasecmp(EntryType, "image/", 6) == 0 &&
+             !g_strncasecmp(DetectedType,MimeTypes[i].str,MimeTypes[i].len)){
+      /* An image, and there's an exact match */
+      st = 0;
+   } else if (g_strncasecmp(EntryType, "text/", 5) ||
+              g_strncasecmp(DetectedType, "application/", 12)) {
+      /* Not an application sent as text */
+      st = 0;
+   }
+
+   return st;
+} 
+
 /*
  * Parse a geometry string.
  */
@@ -296,3 +348,40 @@ gint a_Misc_parse_geometry(gchar *str, gint *x, gint *y, gint *w, gint *h)
    return ret;
 }
 
+/*
+ * Encodes string using base64 encoding.
+ * Return value: new string or NULL if input string is empty.
+ */
+gchar *a_Misc_encode_base64(const gchar *in)
+{
+   static const char *base64_hex = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+                                   "abcdefghijklmnopqrstuvwxyz"
+                                   "0123456789+/";
+   gchar *out = NULL;
+   int len, i = 0;
+
+   if (in == NULL) return NULL;
+   len = strlen(in);
+
+   out = (gchar *)g_malloc((len + 2) / 3 * 4 + 1);
+
+   for (; len >= 3; len -= 3) {
+      out[i++] = base64_hex[in[0] >> 2];
+      out[i++] = base64_hex[((in[0]<<4) & 0x30) | (in[1]>>4)];
+      out[i++] = base64_hex[((in[1]<<2) & 0x3c) | (in[2]>>6)];
+      out[i++] = base64_hex[in[2] & 0x3f];
+      in += 3;
+   }
+
+   if (len > 0) {
+      unsigned char fragment;
+      out[i++] = base64_hex[in[0] >> 2];
+      fragment = (in[0] << 4) & 0x30;
+      if (len > 1) fragment |= in[1] >> 4;
+      out[i++] = base64_hex[fragment];
+      out[i++] = (len < 2) ? '=' : base64_hex[(in[1] << 2) & 0x3c];
+      out[i++] = '=';
+   }
+   out[i] = '\0';
+   return out;
+}
