@@ -39,7 +39,7 @@
 
 #include "url.h"
 
-//#define DEBUG_LEVEL 2
+/*#define DEBUG_LEVEL 2 */
 #include "debug.h"
 
 
@@ -87,7 +87,7 @@ const gchar *a_Url_hostname(const DilloUrl *u)
    if (!url->hostname && url->authority) {
       if ((p = strchr(url->authority, ':'))) {
          url->port = strtol(p + 1, NULL, 10);
-         url->hostname = g_strndup(url->authority, p - url->authority);
+         url->hostname = g_strndup(url->authority,(guint)(p - url->authority));
       } else
          url->hostname = url->authority;
    }
@@ -120,11 +120,11 @@ static DilloUrl *Url_object_new(const gchar *uri_str)
       s = ++p;
    }
    /* p = strpbrk(s, "/"); */
-   if (p && p[0] == '/' && p[1] == '/') {          /* authority */
+   if (p == s && p[0] == '/' && p[1] == '/') {     /* authority */
       s = p + 2;
       p = strpbrk(s, "/?#");
       if (p) {
-         memmove(s - 2, s, MAX(p - s, 1));
+         memmove(s - 2, s, (size_t)MAX(p - s, 1));
          url->authority = s - 2;
          p[-2] = 0;
          s = p;
@@ -186,7 +186,7 @@ static GString *Url_resolve_relative(const gchar *RelStr,
    gchar *p, *s, *e;
    gint i;
    GString *SolvedUrl, *Path;
-   DilloUrl *RelUrl = NULL, *BaseUrl = NULL;
+   DilloUrl *RelUrl, *BaseUrl = NULL;
 
    /* parse relative URL */
    RelUrl = Url_object_new(RelStr);
@@ -219,7 +219,7 @@ static GString *Url_resolve_relative(const gchar *RelStr,
       goto done;
 
    } else if (RelUrl->authority) {               /* authority */
-      // Set the Path buffer and goto "STEP 7";
+      /* Set the Path buffer and goto "STEP 7"; */
       if (RelUrl->path)
          g_string_append(Path, RelUrl->path);
 
@@ -227,7 +227,7 @@ static GString *Url_resolve_relative(const gchar *RelStr,
       g_string_append(Path, RelUrl->path);
 
    } else {
-      // solve relative path
+      /* solve relative path */
       if (BaseUrl->path) {
          g_string_append(Path, BaseUrl->path);
          for (i = Path->len; --i >= 0 && Path->str[i] != '/'; );
@@ -237,19 +237,19 @@ static GString *Url_resolve_relative(const gchar *RelStr,
       if (RelUrl->path)
          g_string_append(Path, RelUrl->path);
 
-      // erase "./"
+      /* erase "./" */
       while ((p=strstr(Path->str, "./")) &&
              (p == Path->str || p[-1] == '/'))
          g_string_erase(Path, p - Path->str, 2);
-      // erase last "."
+      /* erase last "." */
       if (Path->len && Path->str[Path->len - 1] == '.' &&
           (Path->len == 1 || Path->str[Path->len - 2] == '/'))
          g_string_truncate(Path, Path->len - 1);
 
-      // erase "<segment>/../" and "<segment>/.."
+      /* erase "<segment>/../" and "<segment>/.." */
       s = p = Path->str;
       while ( (p = strstr(p, "/..")) != NULL ) {
-         if ((p[3] == '/' || !p[3]) && (p - s)) { //  "/../" | "/.."
+         if ((p[3] == '/' || !p[3]) && (p - s)) { /*  "/../" | "/.."  */
 
             for (e = p + 3 ; p[-1] != '/' && p > s; --p);
             if (p[0] != '.' || p[1] != '.' || p[2] != '/') {
@@ -333,19 +333,28 @@ DilloUrl* a_Url_new(const gchar *url_str, const gchar *base_url,
                     gint flags, gint32 posx, gint32 posy)
 {
    DilloUrl *url;
-   gchar *urlstring, *p, *s, *new_url = NULL;
+   gchar *urlstring, *p, *new_str = NULL;
    GString *SolvedUrl;
-   gint n_ic;
+   gint n_ic, n_ic_spc;
 
    g_return_val_if_fail (url_str != NULL, NULL);
 
-   /* denounce illegal characters (0x00-0x1F, 0x7F and space) */
+   /* Count illegal characters (0x00-0x1F, 0x7F and space) */
    urlstring = (gchar *)url_str;
-   for (p = s = urlstring; *s; s++)
-      if (*s > 0x1F && *s != 0x7F && *s != ' ')
-         p++;
-   /* number of illegal characters */
-   n_ic = s - p;
+   n_ic = n_ic_spc = 0;
+   for (p = urlstring; *p; p++) {
+      n_ic_spc += (*p == ' ') ? 1 : 0;
+      n_ic += (*p != ' ' && *p > 0x1F && *p != 0x7F) ? 0 : 1;
+   }
+   if (n_ic) {
+      /* Strip illegal characters (they could also be encoded).
+       * There's no standard for illegal chars; we chose to strip. */
+      for (p = new_str = g_strdup(urlstring); *urlstring; urlstring++)
+         if (*urlstring > 0x1F && *urlstring != 0x7F && *urlstring != ' ')
+            *p++ = *urlstring;
+      *p = 0;
+      urlstring = new_str;
+   }
 
    /* let's use a heuristic to set http: as default */
    if (!base_url) {
@@ -353,15 +362,14 @@ DilloUrl* a_Url_new(const gchar *url_str, const gchar *base_url,
       if (urlstring[0] != '/') {
          p = strpbrk(urlstring, "/#?:");
          if (!p || *p != ':')
-            urlstring = new_url = g_strconcat("//", urlstring, NULL);
+            urlstring = g_strconcat("//", urlstring, NULL);
       } else if (urlstring[1] != '/')
-         urlstring = new_url = g_strconcat("/", urlstring, NULL);
+         urlstring = g_strconcat("/", urlstring, NULL);
    }
 
    /* Resolve the URL */
    SolvedUrl = Url_resolve_relative(urlstring, NULL, base_url);
    DEBUG_MSG(2, "SolvedUrl = %s\n", SolvedUrl->str);
-   g_free(new_url);
    g_return_val_if_fail (SolvedUrl != NULL, NULL);
 
    /* Fill url data */
@@ -371,7 +379,9 @@ DilloUrl* a_Url_new(const gchar *url_str, const gchar *base_url,
    url->scrolling_position_x = posx;
    url->scrolling_position_y = posy;
    url->illegal_chars = n_ic;
+   url->illegal_chars_spc = n_ic_spc;
 
+   g_free(new_str);
    return url;
 }
 
@@ -394,6 +404,8 @@ DilloUrl* a_Url_dup(const DilloUrl *ori)
    url->ismap_url_len        = ori->ismap_url_len;
    url->scrolling_position_x = ori->scrolling_position_x;
    url->scrolling_position_y = ori->scrolling_position_y;
+   url->illegal_chars        = ori->illegal_chars;
+   url->illegal_chars_spc    = ori->illegal_chars_spc;
 
    return url;
 }
@@ -447,8 +459,7 @@ void a_Url_set_data(DilloUrl *u, gchar *data)
 void a_Url_set_alt(DilloUrl *u, const gchar *alt)
 {
    if (u) {
-      if (u->alt)
-         g_free((gchar *)u->alt);
+      g_free((gchar *)u->alt);
       u->alt = g_strdup(alt);
    }
 }
@@ -485,15 +496,15 @@ void a_Url_set_ismap_coords(DilloUrl *u, gchar *coord_str)
 }
 
 /*
- * Given an hex octet (e3, 2F, 20), return the corresponding
+ * Given an hex octet (e.g., e3, 2F, 20), return the corresponding
  * character if the octet is valid, and -1 otherwise
  */
-static int Url_parse_hex_octet(const gchar *s)
+static int Url_decode_hex_octet(const gchar *s)
 {
    gint hex_value;
    gchar *tail, hex[3];
 
-   if ( (hex[0] = s[0]) && (hex[1] = s[1]) ) {
+   if (s && (hex[0] = s[0]) && (hex[1] = s[1])) {
       hex[2] = 0;
       hex_value = strtol(hex, &tail, 16);
       if (tail - hex == 2)
@@ -503,41 +514,84 @@ static int Url_parse_hex_octet(const gchar *s)
 }
 
 /*
- * Parse possible hexadecimal octets in the URI path
- * Returns new allocated string.
+ * Parse possible hexadecimal octets in the URI path.
+ * Returns a new allocated string.
  */
-gchar *a_Url_parse_hex_path(const DilloUrl *u)
+gchar *a_Url_decode_hex_str(const gchar *str)
 {
-   const gchar *src;
-   gchar *new_uri, *dest;
+   gchar *new_str, *dest;
    int i, val;
 
-   if (!u || !u->path || !u->path[0])
+   if (!str)
       return NULL;
 
    /* most cases won't have hex octets */
-   if (!strchr(u->path, '%'))
-      return g_strdup(u->path);
+   if (!strchr(str, '%'))
+      return g_strdup(str);
 
-   src = u->path;
-   dest = new_uri = g_new(gchar, strlen(src) + 1);
+   dest = new_str = g_new(gchar, strlen(str) + 1);
 
-   for (i = 0; src[i]; i++) {
-      *dest++ = (src[i] == '%' && (val = Url_parse_hex_octet(src+i+1)) >= 0) ?
-                val + !(i+=2) : src[i];
+   for (i = 0; str[i]; i++) {
+      *dest++ = (str[i] == '%' && (val = Url_decode_hex_octet(str+i+1)) >= 0) ?
+                i+=2, val : str[i];
    }
    *dest++ = 0;
 
-   new_uri = g_realloc(new_uri, sizeof(gchar) * (dest - new_uri));
-   return new_uri;
+   new_str = g_realloc(new_str, sizeof(gchar) * (dest - new_str));
+   return new_str;
 }
 
+/*
+ * Urlencode 'str'
+ * -RL :: According to the RFC 1738, only alphanumerics, the special
+ *        characters "$-_.+!*'(),", and reserved characters ";/?:@=&" used
+ *        for their *reserved purposes* may be used unencoded within a URL.
+ * We'll escape everything but alphanumeric and "-_.*" (as lynx).  --Jcid
+ *
+ * Note: the content type "application/x-www-form-urlencoded" is used:
+ *       i.e., ' ' -> '+' and '\n' -> CR LF (see HTML 4.01, Sec. 17.13.4)
+ */
+gchar *a_Url_encode_hex_str(const gchar *str)
+{
+   static const char *verbatim = "-_.*";
+   static const char *hex = "0123456789ABCDEF";
+   char *newstr, *c;
+
+   if (!str)
+      return NULL;
+
+   newstr = g_new(char, 6*strlen(str)+1);
+
+   for (c = newstr; *str; str++)
+      if ((isalnum(*str) && !(*str & 0x80)) || strchr(verbatim, *str))
+      /* we really need isalnum for the "C" locale */
+         *c++ = *str;
+      else if (*str == ' ')
+         *c++ = '+';
+      else if (*str == '\n') {
+         *c++ = '%';
+         *c++ = '0';
+         *c++ = 'D';
+         *c++ = '%';
+         *c++ = '0';
+         *c++ = 'A';
+      } else {
+         *c++ = '%';
+         *c++ = hex[(*str >> 4) & 15];
+         *c++ = hex[*str & 15];
+      }
+   *c = 0;
+
+  return newstr;
+}
+
+
 /*
  * RFC-2396 suggests this stripping when "importing" URLs from other media.
  * Strip: "URL:", enclosing < >, and embedded whitespace.
  * (We also strip illegal chars: 00-1F and 7F)
  */
-gchar *a_Url_string_strip_delimiters(gchar *str)
+gchar *a_Url_string_strip_delimiters(const gchar *str)
 {
    gchar *p, *new_str, *text;
 
